﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using JetBrains.Annotations;
using UnityEngine;

namespace Magify
{
    internal class CountersStorage : IDisposable
    {
        [NotNull]
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Storage);

        [NotNull]
        private readonly string _folder;
        [NotNull]
        private readonly Dictionary<CounterType, BinaryStorage> _storages = new();

        public CountersStorage([NotNull] string folder)
        {
            _folder = folder;
        }

        [NotNull]
        public BinaryStorage this[CounterType type]
        {
            get
            {
                if (!_storages.TryGetValue(type, out var storage))
                {
                    storage = BinaryStorage
                        .Construct(Path.Combine(_folder, $"{type.ToEnumString()}.bin"))
                        .AddEntry(BinaryTypeString.Shared)
                        .AddEntry(BinaryTypeInt32.Shared)
                        .SupportDictionariesOf<string, int>()
                        .Build();
                    _storages[type] = storage;
                }
                return storage;
            }
        }

        public void Dispose()
        {
            foreach (var pair in _storages)
            {
                pair.Value.Dispose();
            }
            _storages.Clear();
        }

        /// <returns>Content of all counters as dictionary in JSON format.</returns>
        [NotNull]
        internal string GetFileContent()
        {
            var map = new Dictionary<CounterType, string>(_storages.Count);
            foreach (var (counter, storage) in _storages)
            {
                var content = storage!.GetFileContent();
                map.Add(counter, content);
            }
            return JsonFacade.SerializeObject(map) ?? string.Empty; // ToDo: optimize serialization.
        }

        /// <summary>
        /// Rewrite content of all counters from JSON map.
        /// </summary>
        internal void RewriteContent([NotNull] string countersContentMap)
        {
            var map = JsonFacade.DeserializeObject<Dictionary<CounterType, string>>(countersContentMap); // ToDo: optimize serialization.
            if (map == null || map.Count == 0)
            {
                return;
            }
            foreach (var (counter, storageContent) in map)
            {
                _logger.Log($"Rewriting content of counter storage: {counter}, content: \n{storageContent}");
                this[counter].RewriteContent(storageContent ?? string.Empty);
            }
        }
    }
}