using System.Text;
using System.Threading;
using Magify.Model;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using UnityEngine;
using UnityEngine.Networking;

namespace Magify
{
    internal class NetworkClient : INetworkClient
    {
        [NotNull]
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Network);
        [NotNull]
        private static readonly UTF8Encoding _utf8Encoding = new();

        public async UniTask<WebResponseMessage> SendAsync([NotNull] string url, WebRequestMessage message, CancellationToken cancellationToken)
        {
            try
            {
                _logger.Log($"SendAsync with message => {message}");
                using var request = await CreateRequest(url, message, cancellationToken);
                await request.SendWebRequest().WithCancellation(cancellationToken);
                cancellationToken.ThrowIfCancellationRequested();
                return request.ToWebRequestResponse(message);
            }
            catch (UnityException e)
            {
                _logger.LogError("Failed to send web request with exception." +
                                 "Don't worry, Magify will try to send it bit later again." +
                                 "But you should send info about this error to Magify support." +
                                 $"\n{e}");
                return new WebResponseMessage
                {
                    Result = UnityWebRequest.Result.ConnectionError,
                    ResponseCode = -1,
                    Error = null,
                    Text = null,
                    RequestMessage = message,
                };
            }
        }

        [ItemNotNull]
        private static async UniTask<UnityWebRequest> CreateRequest(
            [NotNull] string url,
            WebRequestMessage message,
            CancellationToken cancellationToken)
        {
            var payload = new JsonRPCPayload(message.Method, message.Payload);
            var payloadString = JsonFacade.SerializeObject(payload);
            var payloadBytes = _utf8Encoding.GetBytes(payloadString);

            UnityWebRequest request;
            try
            {
                request = CreateRequest(url, payloadBytes);
            }
            catch (UnityException) when (!PlayerLoopHelper.IsMainThread)
            {
                await UniTask.SwitchToMainThread(cancellationToken);
                request = CreateRequest(url, payloadBytes);
            }
            request.SetRequestHeader("Content-Type", "application/json");
            if (message.AuthToken != null)
            {
                request.SetRequestHeader("Authorization", message.AuthToken);
            }

            return request;
        }

        [NotNull]
        private static UnityWebRequest CreateRequest([NotNull] string url, [NotNull] byte[] bytes)
        {
            return new UnityWebRequest(url, UnityWebRequest.kHttpVerbPOST)
            {
                uploadHandler = new UploadHandlerRaw(bytes),
                downloadHandler = new DownloadHandlerBuffer(),
                timeout = 35,
            };
        }
    }
}